@@ -649,3 +649,78 @@ soup_hsts_enforcer_has_valid_policy (SoupHSTSEnforcer *hsts_enforcer,
 
 	return retval;
 }
+
+static void
+add_domain_to_list (gpointer key,
+		    gpointer value,
+		    gpointer data)
+{
+	GList **domains = (GList **) data;
+	*domains = g_list_prepend (*domains, g_strdup ((char*)key));
+}
+
+/**
+ * soup_hsts_enforcer_get_domains:
+ * @hsts_enforcer: a #SoupHSTSEnforcer
+ * @session_policies: whether to include session policies
+ *
+ * Gets a list of domains for which there are policies in @enforcer.
+ *
+ * Since: 2.68
+ *
+ * Returns: (element-type utf8) (transfer full): a newly allocated
+ * list of domains. Use g_list_free_full() and g_free() to free the
+ * list.
+ **/
+GList*
+soup_hsts_enforcer_get_domains (SoupHSTSEnforcer *hsts_enforcer,
+				gboolean          session_policies)
+{
+	GList *domains = NULL;
+
+	g_return_val_if_fail (SOUP_IS_HSTS_ENFORCER (hsts_enforcer), NULL);
+
+	g_hash_table_foreach (hsts_enforcer->priv->host_policies, add_domain_to_list, &domains);
+	if (session_policies)
+		g_hash_table_foreach (hsts_enforcer->priv->session_policies, add_domain_to_list, &domains);
+
+	return domains;
+}
+
+static void
+add_policy_to_list (gpointer key,
+		    gpointer value,
+		    gpointer data)
+{
+	GList **policies = (GList **) data;
+	*policies = g_list_prepend (*policies, soup_hsts_policy_copy ((SoupHSTSPolicy*)value));
+}
+
+/**
+ * soup_hsts_enforcer_all_policies:
+ * @hsts_enforcer: a #SoupHSTSEnforcer
+ * @session_policies: whether to include session policies
+ *
+ * Gets a list with the policies in @enforcer.
+ *
+ * Returns: (element-type SoupHSTSPolicy) (transfer full): a newly
+ * allocated list of policies. Use g_list_free_full() and
+ * soup_hsts_policy_free() to free the list.
+ *
+ * Since: 2.68
+ *
+ **/
+GList*
+soup_hsts_enforcer_get_policies (SoupHSTSEnforcer *hsts_enforcer,
+				 gboolean          session_policies)
+{
+	GList *policies = NULL;
+
+	g_return_val_if_fail (SOUP_IS_HSTS_ENFORCER (hsts_enforcer), NULL);
+
+	g_hash_table_foreach (hsts_enforcer->priv->host_policies, add_policy_to_list, &policies);
+	if (session_policies)
+		g_hash_table_foreach (hsts_enforcer->priv->session_policies, add_policy_to_list, &policies);
+
+	return policies;
+}
